@@ -19,8 +19,9 @@ limitations under the License.
 #include "query.hh"
 
 namespace ccls {
-REFLECT_STRUCT(QueryFile::Def, path, args, language, skipped_ranges,
-               dependencies);
+REFLECT_STRUCT(IndexInclude, line, resolved_path);
+REFLECT_STRUCT(QueryFile::Def, path, args, language, dependencies, includes,
+               skipped_ranges);
 
 namespace {
 struct Out_cclsInfo {
@@ -53,7 +54,16 @@ void MessageHandler::ccls_info(EmptyParam &, ReplyOnce &reply) {
   reply(result);
 }
 
-void MessageHandler::ccls_fileInfo(TextDocumentParam &param, ReplyOnce &reply) {
+struct FileInfoParam : TextDocumentParam {
+  bool dependencies = false;
+  bool includes = false;
+  bool skipped_ranges = false;
+};
+REFLECT_STRUCT(FileInfoParam, textDocument, dependencies, includes, skipped_ranges);
+
+void MessageHandler::ccls_fileInfo(JsonReader &reader, ReplyOnce &reply) {
+  FileInfoParam param;
+  Reflect(reader, param);
   QueryFile *file = FindFile(param.textDocument.uri.GetPath());
   if (!file)
     return;
@@ -63,8 +73,12 @@ void MessageHandler::ccls_fileInfo(TextDocumentParam &param, ReplyOnce &reply) {
   result.path = file->def->path;
   result.args = file->def->args;
   result.language = file->def->language;
-  result.includes = file->def->includes;
-  result.skipped_ranges = file->def->skipped_ranges;
+  if (param.dependencies)
+    result.dependencies = file->def->dependencies;
+  if (param.includes)
+    result.includes = file->def->includes;
+  if (param.skipped_ranges)
+    result.skipped_ranges = file->def->skipped_ranges;
   reply(result);
 }
 } // namespace ccls
