@@ -126,14 +126,20 @@ bool CacheInvalid(VFS *vfs, IndexFile *prev, const std::string &path,
 
   // For inferred files, allow -o a a.cc -> -o b b.cc
   std::string stem = sys::path::stem(path);
-  bool changed = prev->args.size() != args.size();
-  for (size_t i = 0; !changed && i < args.size(); i++)
-    if (strcmp(prev->args[i], args[i]) && sys::path::stem(args[i]) != stem)
-      changed = true;
-  if (changed)
+  int changed = -1, size = std::min(prev->args.size(), args.size());
+  for (int i = 0; i < size; i++)
+    if (strcmp(prev->args[i], args[i]) && sys::path::stem(args[i]) != stem) {
+      changed = i;
+      break;
+    }
+  if (changed < 0 && prev->args.size() != args.size())
+    changed = size;
+  if (changed >= 0)
     LOG_V(1) << "args changed for " << path
-             << (from ? " (via " + *from + ")" : std::string());
-  return changed;
+             << (from ? " (via " + *from + ")" : std::string()) << "; old: "
+             << (changed < prev->args.size() ? prev->args[changed] : "")
+             << "; new: " << (changed < size ? args[changed] : "");
+  return changed >= 0;
 };
 
 std::string AppendSerializationFormat(const std::string &base) {
@@ -218,7 +224,6 @@ bool Indexer_Parse(SemaManager *completion, WorkingFiles *wfiles,
     return false;
   }
 
-  // must_exist is currently unused.
   Project::Entry entry =
       project->FindEntry(request.path, true, request.must_exist);
   if (request.must_exist && entry.filename.empty())
@@ -228,25 +233,28 @@ bool Indexer_Parse(SemaManager *completion, WorkingFiles *wfiles,
   std::string path_to_index = entry.filename;
   std::unique_ptr<IndexFile> prev;
 
-  bool deleted = false, no_linkage = g_config->index.initialNoLinkage ||
-                                     request.mode != IndexMode::Background;
+  bool deleted = request.mode == IndexMode::Delete,
+       no_linkage = g_config->index.initialNoLinkage ||
+                    request.mode != IndexMode::Background;
   int reparse = 0;
-  std::optional<int64_t> write_time = LastWriteTime(path_to_index);
-  if (!write_time) {
-    deleted = true;
-  } else {
-    if (vfs->Stamp(path_to_index, *write_time, no_linkage ? 2 : 0))
-      reparse = no_linkage ? 2 : 1;
-    if (request.path != path_to_index) {
-      std::optional<int64_t> mtime1 = LastWriteTime(request.path);
-      if (!mtime1)
-        deleted = true;
-      else if (vfs->Stamp(request.path, *mtime1, no_linkage ? 2 : 0))
-        reparse = 2;
-    }
-  }
   if (deleted)
     reparse = 2;
+  else if (!(g_config->index.onChange && wfiles->GetFile(path_to_index))) {
+    std::optional<int64_t> write_time = LastWriteTime(path_to_index);
+    if (!write_time) {
+      deleted = true;
+    } else {
+      if (vfs->Stamp(path_to_index, *write_time, no_linkage ? 2 : 0))
+        reparse = no_linkage ? 2 : 1;
+      if (request.path != path_to_index) {
+        std::optional<int64_t> mtime1 = LastWriteTime(request.path);
+        if (!mtime1)
+          deleted = true;
+        else if (vfs->Stamp(request.path, *mtime1, no_linkage ? 2 : 0))
+          reparse = 2;
+      }
+    }
+  }
 
   if (g_config->index.onChange) {
     reparse = 2;
@@ -804,11 +812,10 @@ static void Reply(RequestId id, const char *key,
     w.Null();
     break;
   case RequestId::kInt:
-    w.Int(id.value);
+    w.Int(atoll(id.value.c_str()));
     break;
   case RequestId::kString:
-    auto s = std::to_string(id.value);
-    w.String(s.c_str(), s.length());
+    w.String(id.value.c_str(), id.value.size());
     break;
   }
   w.Key(key);
