@@ -60,7 +60,8 @@ std::pair<LanguageId, bool> lookupExtension(std::string_view filename) {
   bool objc = types::isObjC(I);
   LanguageId ret;
   if (types::isCXX(I))
-    ret = objc ? LanguageId::ObjCpp : LanguageId::Cpp;
+    ret = types::isCuda(I) ? LanguageId::Cuda
+                           : objc ? LanguageId::ObjCpp : LanguageId::Cpp;
   else if (objc)
     ret = LanguageId::ObjC;
   else if (I == types::TY_C || I == types::TY_CHeader)
@@ -95,7 +96,12 @@ struct ProjectProcessor {
         LOG_S(WARNING) << toString(glob_or_err.takeError());
   }
 
-  bool ExcludesArg(StringRef arg) {
+  bool ExcludesArg(StringRef arg, int &i) {
+    if (arg.startswith("-M")) {
+      if (arg == "-MF" || arg == "-MT" || arg == "-MQ")
+        i++;
+      return true;
+    }
     return exclude_args.count(arg) || any_of(exclude_globs,
       [&](const GlobPattern &glob) { return glob.match(arg); });
   }
@@ -117,6 +123,8 @@ struct ProjectProcessor {
             ok |= lang == LanguageId::C && header;
           else if (A.consume_front("%cpp "))
             ok |= lang == LanguageId::Cpp;
+          else if (A.consume_front("%cu "))
+            ok |= lang == LanguageId::Cuda;
           else if (A.consume_front("%hpp "))
             ok |= lang == LanguageId::Cpp && header;
           else if (A.consume_front("%objective-c "))
@@ -128,7 +136,7 @@ struct ProjectProcessor {
         }
         if (ok)
           args.push_back(A.data());
-      } else if (!ExcludesArg(A)) {
+      } else if (!ExcludesArg(A, i)) {
         args.push_back(arg);
       }
     }
@@ -240,8 +248,6 @@ std::vector<const char *> GetFallback(const std::string path) {
   std::vector<const char *> argv{"clang"};
   if (sys::path::extension(path) == ".h")
     argv.push_back("-xobjective-c++-header");
-  for (const std::string &arg : g_config->clang.extraArgs)
-    argv.push_back(Intern(arg));
   argv.push_back(Intern(path));
   return argv;
 }
@@ -273,7 +279,7 @@ void LoadDirectoryListing(ProjectProcessor &proc, const std::string &root,
                          files.push_back(path);
                      } else if (sys::path::filename(path) == ".ccls") {
                        std::vector<const char *> args = ReadCompilerArgumentsFromFile(path);
-                       folder.dot_ccls.emplace(sys::path::parent_path(path),
+                       folder.dot_ccls.emplace(sys::path::parent_path(path).str() + '/',
                                                args);
                        std::string l;
                        for (size_t i = 0; i < args.size(); i++) {
@@ -312,15 +318,40 @@ void LoadDirectoryListing(ProjectProcessor &proc, const std::string &root,
 // Computes a score based on how well |a| and |b| match. This is used for
 // argument guessing.
 int ComputeGuessScore(std::string_view a, std::string_view b) {
-  // Increase score based on common prefix and suffix. Prefixes are prioritized.
-  if (a.size() > b.size())
-    std::swap(a, b);
-  size_t i = std::mismatch(a.begin(), a.end(), b.begin()).first - a.begin();
-  size_t j = std::mismatch(a.rbegin(), a.rend(), b.rbegin()).first - a.rbegin();
-  int score = 10 * i + j;
-  if (i + j < a.size())
-    score -= 100 * (std::count(a.begin() + i, a.end() - j, '/') +
-                    std::count(b.begin() + i, b.end() - j, '/'));
+  int score = 0;
+  unsigned h = 0;
+  llvm::SmallDenseMap<unsigned, int> m;
+  for (uint8_t c : a)
+    if (c == '/') {
+      score -= 9;
+      if (h)
+        m[h]++;
+      h = 0;
+    } else {
+      h = h * 33 + c;
+    }
+  h = 0;
+  for (uint8_t c : b)
+    if (c == '/') {
+      score -= 9;
+      auto it = m.find(h);
+      if (it != m.end() && it->second > 0) {
+        it->second--;
+        score += 31;
+      }
+      h = 0;
+    } else {
+      h = h * 33 + c;
+    }
+
+  uint8_t c;
+  int d[127] = {};
+  for (int i = a.size(); i-- && (c = a[i]) != '/'; )
+    if (c < 127)
+      d[c]++;
+  for (int i = b.size(); i-- && (c = b[i]) != '/'; )
+    if (c < 127 && d[c])
+      d[c]--, score++;
   return score;
 }
 
@@ -332,8 +363,12 @@ void Project::LoadDirectory(const std::string &root, Project::Folder &folder) {
   folder.entries.clear();
   if (g_config->compilationDatabaseCommand.empty()) {
     CDBDir = root;
-    if (g_config->compilationDatabaseDirectory.size())
-      sys::path::append(CDBDir, g_config->compilationDatabaseDirectory);
+    if (g_config->compilationDatabaseDirectory.size()) {
+      if (sys::path::is_absolute(g_config->compilationDatabaseDirectory))
+        CDBDir = g_config->compilationDatabaseDirectory;
+      else
+        sys::path::append(CDBDir, g_config->compilationDatabaseDirectory);
+    }
     sys::path::append(Path, CDBDir, "compile_commands.json");
   } else {
     // If `compilationDatabaseCommand` is specified, execute it to get the
@@ -413,10 +448,10 @@ void Project::LoadDirectory(const std::string &root, Project::Folder &folder) {
 
       std::vector<std::string> args = std::move(Cmd.CommandLine);
       entry.args.reserve(args.size());
-      for (std::string &arg : args) {
-        DoPathMapping(arg);
-        if (!proc.ExcludesArg(arg))
-          entry.args.push_back(Intern(arg));
+      for (int i = 0; i < args.size(); i++) {
+        DoPathMapping(args[i]);
+        if (!proc.ExcludesArg(args[i], i))
+          entry.args.push_back(Intern(args[i]));
       }
       entry.compdb_size = entry.args.size();
 
@@ -442,13 +477,6 @@ void Project::LoadDirectory(const std::string &root, Project::Folder &folder) {
   sys::path::append(Path, root, ".ccls");
   if (sys::fs::exists(Path))
     LoadDirectoryListing(proc, root, Seen);
-  std::vector<const char *> extra_args;
-  for (const std::string &arg : g_config->clang.extraArgs)
-    extra_args.push_back(Intern(arg));
-  for (auto &e : folder.entries) {
-    e.args.insert(e.args.end(), extra_args.begin(), extra_args.end());
-    e.args.push_back(Intern("-working-directory=" + e.directory));
-  }
 }
 
 void Project::Load(const std::string &root) {
@@ -470,94 +498,106 @@ void Project::Load(const std::string &root) {
 
 Project::Entry Project::FindEntry(const std::string &path, bool can_redirect,
                                   bool must_exist) {
-  Project::Folder *best_folder = nullptr;
+  std::string best_dot_ccls_root;
+  Project::Folder *best_dot_ccls_folder = nullptr;
+  std::string best_dot_ccls_dir;
+  const std::vector<const char *> *best_dot_ccls_args = nullptr;
+
+  bool match = false, exact_match = false;
   const Entry *best = nullptr;
-  std::lock_guard lock(mtx);
-  for (auto &[root, folder] : root2folder) {
-    // The entry may have different filename but it doesn't matter when building
-    // CompilerInvocation. The main filename is specified separately.
-    auto it = folder.path2entry_index.find(path);
-    if (it != folder.path2entry_index.end()) {
-      Project::Entry &entry = folder.entries[it->second];
-      if (can_redirect || entry.filename == path)
-        return entry;
-      if (entry.compdb_size) {
-        best_folder = &folder;
-        best = &entry;
-      }
-      break;
-    }
-  }
+  Project::Folder *best_compdb_folder = nullptr;
 
-  bool exists = false;
-  std::string dir;
-  const std::vector<const char *> *extra = nullptr;
   Project::Entry ret;
+  std::lock_guard lock(mtx);
+
   for (auto &[root, folder] : root2folder)
-    if (StringRef(path).startswith(root))
-      for (auto &[dir1, args] : folder.dot_ccls)
-        if (StringRef(path).startswith(dir1)) {
-          dir = dir1;
-          extra = &args;
-          if (AppendToCDB(args))
-            goto out;
-          exists = true;
-
-          ret.root = ret.directory = root;
-          ret.filename = path;
-          if (args.empty()) {
-            ret.args = GetFallback(path);
-          } else {
-            ret.args = args;
-            ret.args.push_back(Intern(path));
-          }
-          ProjectProcessor(folder).Process(ret);
-          for (const std::string &arg : g_config->clang.extraArgs)
-            ret.args.push_back(Intern(arg));
-          ret.args.push_back(Intern("-working-directory=" + ret.directory));
-          return ret;
+    if (StringRef(path).startswith(root)) {
+      // Find the best-fit .ccls
+      for (auto &[dir, args] : folder.dot_ccls)
+        if (StringRef(path).startswith(dir) &&
+            dir.length() > best_dot_ccls_dir.length()) {
+          best_dot_ccls_root = root;
+          best_dot_ccls_folder = &folder;
+          best_dot_ccls_dir = dir;
+          best_dot_ccls_args = &args;
         }
-out:
-  if (must_exist && !exists)
-    return ret;
 
-  if (!best) {
-    int best_score = INT_MIN;
-    for (auto &[root, folder] : root2folder) {
-      if (dir.size() && !StringRef(path).startswith(dir))
-        continue;
-      for (const Entry &e : folder.entries)
-        if (e.compdb_size) {
-          int score = ComputeGuessScore(path, e.filename);
-          if (score > best_score) {
-            best_score = score;
-            best = &e;
-            best_folder = &folder;
+      if (!match) {
+        auto it = folder.path2entry_index.find(path);
+        if (it != folder.path2entry_index.end()) {
+          Project::Entry &entry = folder.entries[it->second];
+          exact_match = entry.filename == path;
+          if ((match = exact_match || can_redirect) || entry.compdb_size) {
+            // best->compdb_size is >0 for a compdb entry, 0 for a .ccls entry.
+            best_compdb_folder = &folder;
+            best = &entry;
           }
         }
+      }
     }
-  }
 
-  ret.is_inferred = true;
-  ret.filename = path;
-  if (!best) {
-    if (ret.root.empty())
-      ret.root = g_config->fallbackFolder;
-    ret.directory = ret.root;
-    ret.args = GetFallback(path);
+  bool append;
+  if (best_dot_ccls_args && !(append = AppendToCDB(*best_dot_ccls_args)) &&
+      !exact_match) {
+    // If the first line is not %compile_commands.json, override the compdb
+    // match if it is not an exact match.
+    ret.root = ret.directory = best_dot_ccls_root;
+    ret.filename = path;
+    if (best_dot_ccls_args->empty()) {
+      ret.args = GetFallback(path);
+    } else {
+      ret.args = *best_dot_ccls_args;
+      ret.args.push_back(Intern(path));
+    }
   } else {
-    ret.root = best->root;
-    ret.directory = best->directory;
-    ret.args = best->args;
-    ret.args.resize(best->compdb_size);
-    if (extra && extra->size())
-      ret.args.insert(ret.args.end(), extra->begin() + 1, extra->end());
-    ProjectProcessor(*best_folder).Process(ret);
-    for (const std::string &arg : g_config->clang.extraArgs)
-      ret.args.push_back(Intern(arg));
-    ret.args.push_back(Intern("-working-directory=" + ret.directory));
+    // If the first line is %compile_commands.json, find the matching compdb
+    // entry and append .ccls args.
+    if (must_exist && !match && !(best_dot_ccls_args && !append))
+      return ret;
+    if (!best) {
+      // Infer args from a similar path.
+      int best_score = INT_MIN;
+      for (auto &[root, folder] : root2folder)
+        if (StringRef(path).startswith(root))
+          for (const Entry &e : folder.entries)
+            if (e.compdb_size) {
+              int score = ComputeGuessScore(path, e.filename);
+              if (score > best_score) {
+                best_score = score;
+                best_compdb_folder = &folder;
+                best = &e;
+              }
+            }
+      ret.is_inferred = true;
+    }
+    if (!best) {
+      ret.root = ret.directory = g_config->fallbackFolder;
+      ret.args = GetFallback(path);
+    } else {
+      // The entry may have different filename but it doesn't matter when
+      // building CompilerInvocation. The main filename is specified
+      // separately.
+      ret.root = best->root;
+      ret.directory = best->directory;
+      ret.args = best->args;
+      if (best->compdb_size) // delete trailing .ccls options if exist
+        ret.args.resize(best->compdb_size);
+      else
+        best_dot_ccls_args = nullptr;
+    }
+    ret.filename = path;
   }
 
+  if (best_dot_ccls_args && append && best_dot_ccls_args->size())
+    ret.args.insert(ret.args.end(), best_dot_ccls_args->begin() + 1,
+                    best_dot_ccls_args->end());
+  if (best_compdb_folder)
+    ProjectProcessor(*best_compdb_folder).Process(ret);
+  else if (best_dot_ccls_folder)
+    ProjectProcessor(*best_dot_ccls_folder).Process(ret);
+  for (const std::string &arg : g_config->clang.extraArgs)
+    ret.args.push_back(Intern(arg));
+  ret.args.push_back(Intern("-working-directory=" + ret.directory));
   return ret;
 }
 
@@ -565,6 +605,9 @@ void Project::Index(WorkingFiles *wfiles, RequestId id) {
   auto &gi = g_config->index;
   GroupMatch match(gi.whitelist, gi.blacklist),
       match_i(gi.initialWhitelist, gi.initialBlacklist);
+  std::vector<const char *> args, extra_args;
+  for (const std::string &arg : g_config->clang.extraArgs)
+    extra_args.push_back(Intern(arg));
   {
     std::lock_guard lock(mtx);
     for (auto &[root, folder] : root2folder) {
@@ -574,7 +617,10 @@ void Project::Index(WorkingFiles *wfiles, RequestId id) {
         if (match.Matches(entry.filename, &reason) &&
             match_i.Matches(entry.filename, &reason)) {
           bool interactive = wfiles->GetFile(entry.filename) != nullptr;
-          pipeline::Index(entry.filename, entry.args,
+          args = entry.args;
+          args.insert(args.end(), extra_args.begin(), extra_args.end());
+          args.push_back(Intern("-working-directory=" + entry.directory));
+          pipeline::Index(entry.filename, args,
                           interactive ? IndexMode::Normal
                                       : IndexMode::Background,
                           false, id);
@@ -597,15 +643,20 @@ void Project::IndexRelated(const std::string &path) {
   auto &gi = g_config->index;
   GroupMatch match(gi.whitelist, gi.blacklist);
   std::string stem = sys::path::stem(path);
+  std::vector<const char *> args, extra_args;
+  for (const std::string &arg : g_config->clang.extraArgs)
+    extra_args.push_back(Intern(arg));
   std::lock_guard lock(mtx);
   for (auto &[root, folder] : root2folder)
     if (StringRef(path).startswith(root)) {
       for (const Project::Entry &entry : folder.entries) {
         std::string reason;
+        args = entry.args;
+        args.insert(args.end(), extra_args.begin(), extra_args.end());
+        args.push_back(Intern("-working-directory=" + entry.directory));
         if (sys::path::stem(entry.filename) == stem && entry.filename != path &&
             match.Matches(entry.filename, &reason))
-          pipeline::Index(entry.filename, entry.args, IndexMode::Background,
-                          true);
+          pipeline::Index(entry.filename, args, IndexMode::Background, true);
       }
       break;
     }
