@@ -76,7 +76,7 @@ TextEdit ToTextEdit(const clang::SourceManager &SM, const clang::LangOptions &L,
                     const clang::FixItHint &FixIt) {
   TextEdit edit;
   edit.newText = FixIt.CodeToInsert;
-  auto r = FromCharSourceRange(SM, L, FixIt.RemoveRange);
+  auto r = fromCharSourceRange(SM, L, FixIt.RemoveRange);
   edit.range =
       lsRange{{r.start.line, r.start.column}, {r.end.line, r.end.column}};
   return edit;
@@ -194,7 +194,7 @@ public:
                           SrcMgr::CharacteristicKind FileKind) override {
     (void)SM;
     if (File && Seen.insert(File).second)
-      out.emplace_back(PathFromFileEntry(*File), File->getModificationTime());
+      out.emplace_back(pathFromFileEntry(*File), File->getModificationTime());
   }
 };
 
@@ -238,7 +238,7 @@ public:
     auto it = FID2concerned.try_emplace(FID.getHashValue());
     if (it.second) {
       const FileEntry *FE = SM.getFileEntryForID(FID);
-      it.first->second = FE && PathFromFileEntry(*FE) == path;
+      it.first->second = FE && pathFromFileEntry(*FE) == path;
     }
     return it.first->second;
   }
@@ -260,7 +260,7 @@ public:
       llvm::SmallString<64> Message;
       Info.FormatDiagnostic(Message);
       d.range =
-          FromCharSourceRange(SM, *LangOpts, DiagnosticRange(Info, *LangOpts));
+          fromCharSourceRange(SM, *LangOpts, DiagnosticRange(Info, *LangOpts));
       d.message = Message.str();
       d.concerned = concerned;
       d.file = Filename;
@@ -285,6 +285,8 @@ public:
       } else {
         Note &n = last->notes.emplace_back();
         fillDiagBase(n);
+        if (concerned)
+          last->concerned = true;
       }
     } else {
       Flush();
@@ -313,6 +315,7 @@ std::unique_ptr<CompilerInstance> BuildCompilerInstance(
       Clang->getDiagnostics(), Clang->getInvocation().TargetOpts));
   if (!Clang->hasTarget())
     return nullptr;
+  Clang->getPreprocessorOpts().RetainRemappedFileBuffers = true;
   // Construct SourceManager with UserFilesAreVolatile: true because otherwise
   // RequiresNullTerminator: true may cause out-of-bounds read when a file is
   // mmap'ed but is saved concurrently.
@@ -336,8 +339,15 @@ bool Parse(CompilerInstance &Clang) {
   SyntaxOnlyAction Action;
   if (!Action.BeginSourceFile(Clang, Clang.getFrontendOpts().Inputs[0]))
     return false;
+#if LLVM_VERSION_MAJOR >= 9 // rL364464
+  if (llvm::Error E = Action.Execute()) {
+    llvm::consumeError(std::move(E));
+    return false;
+  }
+#else
   if (!Action.Execute())
     return false;
+#endif
   Action.EndSourceFile();
   return true;
 }
@@ -362,6 +372,7 @@ void BuildPreamble(Session &session, CompilerInvocation &CI,
   CI.getDiagnosticOpts().IgnoreWarnings = false;
   CI.getFrontendOpts().SkipFunctionBodies = true;
   CI.getLangOpts()->CommentOpts.ParseAllComments = g_config->index.comments > 1;
+  CI.getLangOpts()->RetainCommentsFromSystemHeaders = true;
 
   StoreDiags DC(task.path);
   IntrusiveRefCntPtr<DiagnosticsEngine> DE =
@@ -493,7 +504,6 @@ void *CompletionMain(void *manager_) {
     Clang->setCodeCompletionConsumer(task->Consumer.release());
     if (!Parse(*Clang))
       continue;
-    Buf.release();
 
     task->on_complete(&Clang->getCodeCompletionConsumer());
   }
@@ -583,7 +593,6 @@ void *DiagnosticMain(void *manager_) {
       continue;
     if (!Parse(*Clang))
       continue;
-    Buf.release();
 
     auto Fill = [](const DiagBase &d, Diagnostic &ret) {
       ret.range = lsRange{{d.range.start.line, d.range.start.column},
